# Copyright (c) 2021, NakaMetPy Develoers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
# 
# Function load_jmara_grib2 is based on Qiita article
# URL: https://qiita.com/vpcf/items/b680f504cfe8b6a64222
#

import struct
import numpy as np
from itertools import repeat


def _set_table(section5):
  max_level = struct.unpack_from('>H', section5, 15)[0]
  table = (
    -10, # define representative of level 0　(Missing Value)
    *struct.unpack_from('>'+str(max_level)+'H', section5, 18)
  )
  return np.array(table, dtype=np.int16)

def _decode_runlength(code, hi_level):
  for raw in code:
    if raw <= hi_level:
      level = raw
      pwr = 0
      yield level
    else:
      length = (0xFF - hi_level)**pwr * (raw - (hi_level + 1))
      pwr += 1
      yield from repeat(level, length)

def load_jmara_grib2(file):
  r'''気象庁解析雨量やレーダー雨量を返す関数

  欠損値は負の値として表現される

  Parameters
  --------
  file: `str`
    file path 
    ファイルのPATH

  Returns
  -------
  rain: `numpy.ma.MaskedArray`
    単位 (mm)

  Note
  -----
  ``jma_rain_lat`` , ``jma_rain_lon`` はそれぞれ返り値に対応する
  `np.ndarray` 型の緯度経度である。
  '''
  with open(file, 'rb') as f:
    binary = f.read()
  
  len_ = {'sec0':16, 'sec1':21, 'sec3':72, 'sec4':82, 'sec6':6}
  
  end4 = len_['sec0'] + len_['sec1'] + len_['sec3'] + len_['sec4'] - 1
  len_['sec5'] = struct.unpack_from('>I', binary, end4+1)[0]
  section5 = binary[end4:(end4+len_['sec5']+1)]
  power = section5[17]
  # print(power)
  
  end6 = end4 + len_['sec5'] + len_['sec6']
  len_['sec7'] = struct.unpack_from('>I', binary, end6+1)[0]
  section7 = binary[end6:(end6+len_['sec7']+1)]
  
  highest_level = struct.unpack_from('>H', section5, 13)[0]
  level_table = _set_table(section5)
  decoded = np.fromiter(
    _decode_runlength(section7[6:], highest_level), dtype=np.int16
  ).reshape((3360, 2560))
  
  # convert level to representative
  return np.ma.masked_less((level_table[decoded]/(10**power))[::-1, :], 0)

def get_jmara_lat():
  r'''解析雨量の緯度を返す関数

  Returns
  -------
  lat: `numpy.ndarray`
  '''
  return np.linspace(48, 20, 3360, endpoint=False)[::-1] - 1/80/1.5 / 2
    

def get_jmara_lon():
  r'''解析雨量の経度を返す関数

  Returns
  -------
  lon: `numpy.ndarray`
  '''
  return np.linspace(118, 150, 2560, endpoint=False) + 1/80 / 2

def get_gsmap_lat():
  r'''GSMaPの緯度を返す関数

  Returns
  -------
  lat: `numpy.ndarray`
  '''
  return np.arange(-60, 60, 0.1)[::-1] + 0.05
    

def get_gsmap_lon():
  r'''GSMaPの経度を返す関数

  Returns
  -------
  lon: `numpy.ndarray`
  '''
  return np.arange(0, 360, 0.1) + 0.05


def dt_ymdhm(date, opt=1):
  r'''
  datetime.datetime から year, month, day, hour, minute の set を返す関数。
  opt = 1 : string, 0 : int

  Return the set of year, month, day, hour, minute from datetime.datetime.

  Parameters
  ----------
  date: `datetime.datetime`
    datetime
  opt: `int`
    return string or not
  
  Returns
  -------
  `set`
    (year, month, day, hour, minute)
  '''
  if opt == 0:
    return (date.year, date.month, date.day, date.hour, date.minute)
  elif opt == 1:
    return (f"{date.year}", f"{date.month:02}", f"{date.day:02}", f"{date.hour:02}", f"{date.minute:02}")


jma_rain_lat = np.linspace(48, 20, 3360, endpoint=False)[::-1] - 1/80/1.5 / 2
jma_rain_lon = np.linspace(118, 150, 2560, endpoint=False) + 1/80 / 2

gsmap_lat = np.arange(-60, 60, 0.1)[::-1] + 0.05
gsmap_lon = np.arange(0, 360, 0.1) + 0.05

def unit_ms1_knots(ms):
  r"""
  Convert unit m/s into knots.
  
  Parameters
  ----------
  ms: `int`
    Velocity in meter per second.
  
  Returns
  -------
  `kt`
    Velocity in knots.
  """
  return ms*3600/1852

def unit_knots_ms1(kt):
  r"""
  Convert unit knots into m/s.
  
  Parameters
  ----------
  kt: `int`
    Velocity in knots.
  
  Returns
  -------
  `ms`
    Velocity in meter per second.
  """
  return kt*1852/3600

def anom_levels(levs):
  """
  Return minus ans plus levels.

  Parameters
  ----------
  levs: `list`or`np.ndarray`

  Returns
  -------
  anom levels: `np.ndarray`
  
  Examples  
  --------
      >>> levs = [0.5, 1., 2.]
      >>> print(anom_levels(levs))
      [-2.  -1.  -0.5  0.5  1.   2. ]
  
  
  """
  levs = list(set(np.abs(levs)))
  levs.sort()
  return np.array([-i for i in levs[::-1]]+levs)


# Copyright 2022 nakamura_yuki. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import numpy as np
from numpy import ma
from ._error import ExceedZidxError, ExceedTidxError, InvalidTidxError, InvalidZidxError
import re
import os
import glob
import warnings
import datetime

class GrADS:
  """
  A nakametpy.grads `GrADS` is a collection of dimensions and variables. 

  The following class variables are read-only.

  **`dimensions`**: The `dimensions` dictionary maps the names of
  dimensions defined for the `Dataset` to instances of the
  `Dimension` class.

  **`variables`**: The `variables` dictionary maps the names of variables
  defined for this `Dataset` or `Group` to instances of the
  `Variable` class.
  """
  def __init__(self, filename, endian=None, do_squeeze=True):
    """
      **`__init__(self, filename, s2n=True, endian=None, do_squeese=True`**
      
      `GrADS` constructor which is simular to netCDF4-Python `Dataset` constructor.
      
      **`filename`**: Name of GrADS control file to handle GrADS binary data.
      
      **`s2n`**: Data saved direction in ydef.
      
      **`endian`**: Endian of the binary data. `big_endian`, `little_endian` and `native_endian` are available.
      
      **`do_squeeze`**: Whether drop dimension which have only one element. Default is True.
    """
    self.filename = filename
    self.do_squeese = do_squeeze
    self.dimensions = dict()
    
    s2n = True
    b2t = True
    
    with open(self.filename) as f:
      lines = f.readlines()
      
    _sidx = 1E30 # provisional value
    _eidx = None
    _var_idx = []
    _nzdims = []
    
    for idx, line in enumerate(lines):
      if "*" in line[:3]: # check for comment out
        continue
      else:
        if "dset" in line.lower():
          _binname = _get_line_list(line)[1]
          _binname = _replace_template(_binname)
          if "/" == _binname[0]: # is this enough ?
            self.binname = glob.glob(_binname)[0]
          else:
            self.binname = glob.glob(os.path.join(os.path.dirname(self.filename), _binname))[0]
          continue
            
        if "undef" in line.lower():
          self.undef = float(_get_line_list(line)[1])
          continue
          
        if "options" in line.lower(): # So far, endian, yrev, zrev is available. byteswapped is not supported.
          if "big_endian" in line.lower():
            self.endian = "big_endian"
          elif "little_endian" in line.lower():
            self.endian = "little_endian"
          else:
            self.endian = "native_endian"
          if "yrev" in line.lower():
            s2n = False
          if "zrev" in line.lower():
            b2t = False
          continue
          
        if "xdef" in line.lower():
          _xdef = _get_line_list(line)
          self.nx = int(_xdef[1])
          if "linear" in line.lower():
            self.dimensions["xdef"] = Dimension(np.arange(float(_xdef[3]), float(_xdef[3])+float(_xdef[1])*float(_xdef[4]), float(_xdef[4])), "xdef")
          else:
            warnings.warn("Debug Warning: Currently, Only 'LINEAR' is supported in XDEF.")
          continue
          
        if "ydef" in line.lower():
          _ydef = _get_line_list(line)
          self.ny = int(_ydef[1])
          if "linear" in line.lower():
            self.dimensions["ydef"] = Dimension(np.arange(float(_ydef[3]), float(_ydef[3])+float(_ydef[1])*float(_ydef[4]), float(_ydef[4]))[::(-1)**(int(s2n)+1)], "ydef") # if s2n is True -> +1, if False -> -1
          else:
            warnings.warn("Debug Warning: Currently, Only 'LINEAR' is supported in YDEF.")
          continue
          
        if "zdef" in line.lower():
          _zdef = _get_line_list(line)
          if "levels" in line.lower():
            _zdef = [float(_izdef) for _izdef in _zdef[3:int(_zdef[1])+3]]
            self.dimensions["zdef"] = Dimension(np.array(_zdef[::(-1)**(int(b2t)+1)]), "zdef")
          elif "linear" in line.lower():
            self.dimensions["zdef"] = Dimension(np.arange(float(_zdef[3]), float(_zdef[3])+float(_zdef[1])*float(_zdef[4]), float(_zdef[4]))[::(-1)**(int(b2t)+1)], "zdef") # if b2c is True -> +1, if False -> -1
          else:
            warnings.warn("Debug Warning: Currently, Only 'LEVELS' and 'LINEAR' are supported in ZDEF.")
          continue
          
        if "tdef" in line.lower():
          _tdef = _get_line_list(line)
          self.nt = int(_tdef[1])
          if "linear" in line.lower():
            _init_dt = datetime.datetime.strptime(_tdef[3].title(), "%HZ%d%b%Y")
            if _tdef[4][-2:].lower() == "hr":
              self.dimensions["tdef"] = Dimension(np.array([_init_dt + i*datetime.timedelta(hours=int(_tdef[4][:-2])) for i in range(int(_tdef[1]))]), "tdef")
            elif _tdef[4][-2:].lower() == "dy":
              self.dimensions["tdef"] = Dimension(np.array([_init_dt + i*datetime.timedelta(days=int(_tdef[4][:-2])) for i in range(int(_tdef[1]))]), "tdef")
            elif _tdef[4][-2:].lower() == "mo":
              self.dimensions["tdef"] = Dimension(np.array([_init_dt + i*datetime.timedelta(months=int(_tdef[4][:-2])) for i in range(int(_tdef[1]))]), "tdef")
          else:
            warnings.warn("Debug Warning: Currently, Only 'LINEAR' is supported in TDEF.")
          continue

        if bool(re.match("^vars", line.lower())):
          _sidx = idx
          _nvar2 = int(_get_line_list(line)[1])
          continue
        if bool(re.match("^endvars", line.lower())):
          _eidx = idx
          continue
        if (_sidx<idx) and (_eidx==None):
          _var_idx.append(idx)
          _inzdim = int(_get_line_list(line)[1])
          if _inzdim==0:
            _nzdims.append(1)
          else:
            _nzdims.append(_inzdim)
          
    if endian!=None:
      self.endian = endian
    if len(_var_idx)!=_nvar2:
      warnings.warn("The Number of variables is NOT match.")
      
    self.variables = _get_vars(self, lines, _var_idx, _nzdims, do_squeeze)

def _get_vars(self, lines, idx_list, nz_list, do_squeese): # self : grads.GrADS object
  """
  Get the variables info
  """
  variables = dict()
  for inidx, iline_idx in enumerate(idx_list):
    iline = lines[iline_idx]
    _ivar_list = _get_line_list(iline)
    varname = _ivar_list[0]
    idesc = None
    
    if "**" in iline:
      idesc = iline[(iline.index("**") + 1):].replace("\n", "").strip()
      if idesc == " ":
        idesc = None
    elif "*" in iline:
      idesc = iline[iline.index("*"):].replace("\n", "").strip()
      if idesc == " ":
        idesc = None
    else:
      idesc = None
    
    varids = np.cumsum(np.array(nz_list))
    _loop_block = varids[-1]
    varids = list(varids[:-1])
    varids.insert(0,0)
    varids = np.array(varids)
    
    variables[_ivar_list[0]] = Variable(self.binname, varname, varids[inidx],\
                        _loop_block, self.nx, self.ny, nz_list[inidx],\
                        self.nt, self.endian, self.undef, idesc, do_squeese)
  
  return variables      

def _get_line_list(line):
  """
  Get a list of single line.
  """
  return [i for i in re.split("[\s\t\n^,]+", \
          line.replace("\n", "", 1).strip()) if i != ""]
  
def _replace_template(dset):
  """
  Convert specific simbol into ?, which means some ONE letter.
  """
  for i in range(1, 5, 1):
    dset = re.sub(f"%[a-z]{i}", "?"*i, dset)
  return dset
            
class Variable:
  """
  A GrADS `Variable` is used to read 4 bytes direct access binary data.  
  They are analogous to numpy array objects. 
  See `Variable.__init__` for more details.
  """
  def __init__(self, binname, varname, varid, loop_block, nx, ny, nz, nt, endian, undef, desc, do_squeese):
    """
    **`__init__(self, binname, varname, varid, loop_block, nx, ny, nz, nt, endian, undef, desc, do_squeese)`**
    
    `Variable` constructor.
    """
    self._name = varname
    self._undef = undef
    self._binname = binname
    self._varid = varid
    self._loop_block = loop_block
    self._endian = endian
    self._nx = nx
    self._ny = ny
    self._nz = nz
    self._nt = nt
    self._dtype = np.float32
    self._desc = desc
    self._do_squeese = do_squeese

  def __array__(self):
    return self[...]

  def __repr__(self):
    return self.__str__()

  def __getitem__(self, elem):
    """
    Need for slicing.
    """
    return _sel(self._binname, self._name, self._varid, self._loop_block, self._endian,\
      self._nx, self._ny, self._nz, self._nt, self._undef, self._do_squeese)[elem]

  def __str__(self):
    ncdump = [repr(type(self))]
    ncdump.append(f'{self._dtype} {self._name}({self._nt}, {self._nz}, {self._ny}, {self._nx})')
    ncdump.append(f"    description: {self._desc}")
    ncdump.append(f"    _FillValue: {self._undef}")
    return "\n".join(ncdump)

  def sel(self,  zidx=None, tidx=None):
    return _sel(self._binname, self._name, self._varid, self._loop_block, self._endian, self._nx, self._ny, self._nz, self._nt, self._undef, self._do_squeese,  zidx, tidx)

def _sel(binname, name, varid, loop_block, endian, nx, ny, nz, nt, undef, do_squeese,  zidx=None, tidx=None):
  # check for index exceeding
  if tidx==None:
    tloop = range(nt)
  elif isinstance(tidx, int):
    if tidx >= nt:
      raise ExceedTidxError(name, (nt, nz, ny, nx), tidx)
    tloop = [tidx]
  elif isinstance(tidx, (list, tuple, np.ndarray)):
    for _itidx in tidx:
      if _itidx >= nt:
        raise ExceedTidxError(name, (nt, nz, ny, nx), _itidx)
    tloop = tidx
  else:
    raise InvalidTidxError(type(tidx))

  if zidx==None:
    zloop = range(nz)
  elif isinstance(zidx, int):
    if zidx >= nz:
      raise ExceedZidxError(name, (nt, nz, ny, nx), zidx)
    zloop = [zidx]
  elif isinstance(zidx, (list, tuple, np.ndarray)):
    for _izidx in zidx:
      if _izidx >= nz:
        raise ExceedZidxError(name, (nt, nz, ny, nx), _izidx)
    zloop = zidx
  else:
    raise InvalidZidxError(type(zidx))
  
  with open(binname, "rb") as f:
    _data = []
    for _it in tloop:
      if zloop==range(nz):
        f.seek((_it*loop_block+varid)*nx*ny*4, os.SEEK_SET)
        _data.append(ma.masked_equal(ma.masked_array(np.fromfile(f,\
          dtype=_endian2simbole(endian)+"f4", count=nx*ny*nz)), value=undef).reshape(nz, ny, nx))
      else:
        _idata = []
        for _iz in zloop:
          f.seek((varid+_it*loop_block+_iz)*nx*ny*4, os.SEEK_SET)
          _idata.append(ma.masked_equal(ma.masked_array(np.fromfile(f, dtype=_endian2simbole(endian)+"f4", count=nx*ny)), value=undef).reshape(ny, nx))
        _data.append(_idata)

  _data = ma.masked_array(_data)
  if do_squeese:
    return ma.squeeze(_data)
  else:
    return _data


class Dimension:
  """
  A GrADS `Dimension` is used to describe the coordinates of a `Variable`.
  See `Dimension.__init__` for more details.
  """
  def __init__(self, data, var):
    """
    **`__init__(self, data, var)`**
    """
    self.values = data
    self.dim = var

  def __array__(self):
    return self[...]

  def __repr__(self):
    return self.__str__()

  def __getitem__(self, elem):
    """
    Need for slicing.
    """
    return self.values[elem]

  def __str__(self):
    ncdump = [repr(type(self))]
    ncdump.append(f"    variables(dimensions): {self.dim}, ")
    ncdump.append(f"    dimensions(sizes): {len(self.values)}, ")
    ncdump.append(f"    dimensions(data): {self.values}, ")
    return "\n".join(ncdump)

def _endian2simbole(endian):
  """
  Get NumPy's endian simbol from the words.
  """
  if endian.lower() == "big_endian":
    return ">"
  elif endian.lower() == "little_endian":
    return "<"
  elif endian.lower() == "native_endian":
    return "="
  else:
    return ""


# Copyright 2021 nakamura_yuki
# 
from .util import jma_rain_lat, jma_rain_lon

def load_jmara_grib2(file):
  r'''気象庁解析雨量やレーダー雨量を返す関数

  欠損値は負の値として表現される

  Parameters
  --------
  file: `str`
      file path 
      ファイルのPATH

  Returns
  -------
  rain: `numpy.ma.MaskedArray`
      単位 (mm)

  Note
  -----
  The same as util.load_jmara_grib2.
  '''
  from .util import load_jmara_grib2 as _func
  return _func(file)

jma_rain_lat = jma_rain_lat
jma_rain_lon = jma_rain_lon


def get_jmara_lat():
  r'''解析雨量の緯度を返す関数

  Returns
  -------
  lat: `numpy.ndarray`

  Note
  -----
  The same as util.get_jmara_lat.
  '''
  from .util import get_jmara_lat as _func
  return _func()

def get_jmara_lon():
  r'''解析雨量の経度を返す関数

  Returns
  -------
  lon: `numpy.ndarray`
  
  Note
  -----
  The same as util.get_jmara_lon.
  '''
  from .util import get_jmara_lon as _func
  return _func()


# Copyright (c) 2021, NakaMetPy Develoers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
import numpy as np


class MyException(Exception):
    def __init__(self, arg=""):
        self.arg = arg

class MyException2(Exception):
    def __init__(self, arg='', array1=None, array2=None):
        self.arg = arg
        self.array1 = array1
        self.array2 = array2

class MyException3(Exception):
    def __init__(self, *kargs):
        self.kargs = kargs

class NotHaveEnoughDimsError(MyException):
    def __str__(self):
        # print(f"NotHaveEnoughDimsError: 変数 {self.arg} は1次元配列です。この変数は最低でも2次元である必要があります。")
        return (
            f"変数 {self.arg} は1次元配列です。この変数は最低でも2次元である必要があります。\n"+\
            f"The variable {self.arg} is 1D array. This variable must be at least 2D array."
        )

class NotAllowedDxShapeError(MyException2):
    def __str__(self):
        return (
            f"配列 dx の形が適切ではありません。\n{self.arg} の経度緯度方向は({self.array1.shape[-2]}, {self.array1.shape[-1]}), dx の緯度経度方向は({self.array1.shape[-2]}, {self.array1.shape[-1]-1})またはintかfloatの一定値である必要があります。しかし実際には({self.array2.shape[-2]}, {self.array2.shape[-1]})となっています。\n"+\
            f"The shape of array dx is not correct. The shape of \n{self.arg} on longitude is ({self.array1.shape[-2]}, {self.array1.shape[-1]}) and The shape of dx ({self.array1.shape[-2]}, {self.array1.shape[-1]-1}), OR they should be constant of int or float. However in your script, the shape is ({self.array2.shape[-2]}, {self.array2.shape[-1]})"
        )

class NotAllowedDyShapeError(MyException2):
    def __str__(self):
        return (
            f"配列 dx の形が適切ではありません。\n{self.arg} の経度緯度方向は({self.array1.shape[-2]}, {self.array1.shape[-1]}), dy の緯度経度方向は({self.array1.shape[-2]-1}, {self.array1.shape[-1]})またはintかfloatの一定値である必要があります。しかし実際には({self.array2.shape[-2]}, {self.array2.shape[-1]})となっています。\n"+\
            f"The shape of array dy is not correct. The shape of \n{self.arg} on latitude is ({self.array1.shape[-2]}, {self.array1.shape[-1]}) and The shape of dy ({self.array1.shape[-2]}, {self.array1.shape[-1]-1}), OR they should be constant of int or float. However in your script, the shape is ({self.array2.shape[-2]}, {self.array2.shape[-1]})"
        )

class InvalidDxValueError(Exception):
    def __str__(self):
        return (
            f"dx が0以下、または0以下の要素を含んでいます。値は必ず正でなければなりません。\n"+\
            f"dx includes the values which < or <= 0 . The values must be positive."
        )

class InvalidDyValueError(Exception):
    def __str__(self):
        return (
            f"dy が0以下、または0以下の要素を含んでいます。値は必ず正でなければなりません。\n"+\
            f"dy includes the values which < or <= 0 . The values must be positive."
        )

class ExceedTidxError(MyException3):
    def __str__(self):
        if self.kargs[1][0]==1:
          return (
              f"変数 {self.kargs[0]} の shape は {self.kargs[1]} です。 tidx は {self.kargs[2]} が指定されており {self.kargs[1][0]-1} である必要があります。\n"+\
              f"The shape of variable {self.kargs[0]} is {self.kargs[1]}. You set tidx = {self.kargs[2]} but the tidx must be {self.kargs[1][0]-1}."
          )
        else:
          return (
              f"変数 {self.kargs[0]} の shape は {self.kargs[1]} です。 tidx は {self.kargs[2]} が指定されており {self.kargs[1][0]-1} 以下である必要があります。\n"+\
              f"The shape of variable {self.kargs[0]} is {self.kargs[1]}. You set tidx = {self.kargs[2]} but the tidx must be less than or equal {self.kargs[1][0]-1}."
          )

class ExceedZidxError(MyException3):
    def __str__(self):
        if self.kargs[1][1]==1:
          return (
              f"変数 {self.kargs[0]} の shape は {self.kargs[1]} です。 zidx は {self.kargs[2]} が指定されており {self.kargs[1][1]-1} である必要があります。\n"+\
              f"The shape of variable {self.kargs[0]} is {self.kargs[1]}. You set zidx = {self.kargs[2]} but the zidx must be {self.kargs[1][1]-1}."
          )
        else:
          return (
              f"変数 {self.kargs[0]} の shape は {self.kargs[1]} です。 zidx は {self.kargs[2]} が指定されており {self.kargs[1][1]-1} 以下である必要があります。\n"+\
              f"The shape of variable {self.kargs[0]} is {self.kargs[1]}. You set zidx = {self.kargs[2]} but the zidx must be less than or equal {self.kargs[1][1]-1}."
          )

class InvalidTidxError(MyException3):
    def __str__(self):
      return (
          f"tidx は None, int, list, tuple, np.ndarray のいずれかである必要があります。しかし {self.kargs[0]} となっています。\n"+\
          f"tidx is must be None, int, list, tuple or np.ndarray. However, type(tidx) is now {self.kargs[0]}."
      )

class InvalidZidxError(MyException3):
    def __str__(self):
      return (
          f"zidx は None, int, list, tuple, np.ndarray のいずれかである必要があります。しかし {self.kargs[0]} となっています。\n"+\
          f"zidx is must be None, int, list, tuple or np.ndarray. However, type(zidx) is now {self.kargs[0]}."
      )